/*
 * Copyright (c) 2015,2016 Oracle and/or its affiliates. All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL). See the LICENSE file in the root
 * directory for license terms. You may choose either license, or both.
 */

package com.oracle.iot.sample;

import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.os.PowerManager;
import android.os.Process;
import android.os.RemoteCallbackList;
import android.os.RemoteException;
import android.util.Log;
import android.widget.Toast;
import java.io.FileOutputStream;
import java.io.IOException;

import java.security.GeneralSecurityException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.Locale;

import oracle.iot.client.DeviceModel;
import oracle.iot.client.device.Alert;
import oracle.iot.client.device.GatewayDevice;
import oracle.iot.client.device.VirtualDevice;

/**
 * This sample is a gateway that presents multiple simple sensors as virtual
 * devices to the IoT server.
 * <p>
 * The simple sensors are polled every half second and corresponding virtual
 * device is set.
 */
public class GatewayDeviceSampleService extends Service {
    private static final String ERROR_TAG = "IOT_ERROR";
    private static final String MSG_TAG = "IOT";
    private static GatewayDevice gatewayDevice;
    private Context context;

    private int maxMaxThermoThreshold = 70;
    private int minMaxThermoThreshold = 55;
    private int maxMaxHygroThreshold = 95;
    private int minMaxHygroThreshold = 80;

    private int rememberHygroMaxThreshold = 0;
    private int rememberThermoMaxThreshold = 0;

    private boolean messageSent = false;
    private String messageToBeSent = "";
    private boolean stopped = false;

    private Thread workerThread;
    private Thread sensorThread;

    private final static String HUMIDITY_SENSOR_MODEL_URN = "urn:com:oracle:iot:device:humidity_sensor";
    private final static String TEMPERATURE_SENSOR_MODEL_URN = "urn:com:oracle:iot:device:temperature_sensor";
    private static TemperatureSensor temperatureSensor;

    private static HumiditySensor humiditySensor;


    /**
     * This is a list of callbacks that have been registered with the
     * service.  Note that this is package scoped (instead of private) so
     * that it can be accessed more efficiently from inner classes.
     */
    final RemoteCallbackList<IGatewayDeviceSampleServiceCallback> mCallbacks
            = new RemoteCallbackList<IGatewayDeviceSampleServiceCallback>();

    NotificationManager mNM;
    PowerManager.WakeLock wakeLock;


    @Override
    public void onCreate() {
        mNM = (NotificationManager)getSystemService(NOTIFICATION_SERVICE);
        // Display a notification about us starting.
        showNotification(true);
    }

    @Override
    public IBinder onBind(Intent intent) {
        // Select the interface to return.
        if (IGatewayDeviceSampleService.class.getName().equals(intent.getAction())) {
            return mBinder;
        }
        if (IGatewayDeviceSampleServiceSecondary.class.getName().equals(intent.getAction())) {
            return mSecondaryBinder;
        }
        return null;
    }

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        PowerManager powerManager = (PowerManager) getSystemService(POWER_SERVICE);
        wakeLock = powerManager.newWakeLock(PowerManager.PARTIAL_WAKE_LOCK,
                "IoTWakeLock");
        // power management: Keep the CPU on so we can continue to send messages back to the server
        wakeLock.acquire();
        stopped = false;
        context = this;
        Toast.makeText(this, "Service started.", Toast.LENGTH_LONG).show();
        workerThread = new Thread() {
            @Override
            public void run() {
                runSample();
            }
        };
        workerThread.start();

        return START_NOT_STICKY;
    }

    /**
     * The IGatewayDeviceSampleService Interface is defined through IDL
     */
    private final IGatewayDeviceSampleService.Stub mBinder =
            new IGatewayDeviceSampleService.Stub() {
        public void registerCallback(IGatewayDeviceSampleServiceCallback cb) {
            if (cb != null) mCallbacks.register(cb);
        }
        public void unregisterCallback(IGatewayDeviceSampleServiceCallback cb) {
            if (cb != null) mCallbacks.unregister(cb);
        }
    };
    /**
     * A secondary GD Sample Service interface to the service.
     */
    private final IGatewayDeviceSampleServiceSecondary.Stub mSecondaryBinder =
            new IGatewayDeviceSampleServiceSecondary.Stub() {
                public int getPid() {
                    return Process.myPid();
                }
                public String getMessage() {
                    messageSent = true;
                    return messageToBeSent;
                }
                public int getHygroMaxThreshold() {
                    return humiditySensor.getMaxThreshold();
                }
                public void increaseHygroMaxThreshold() {
                    int mt = humiditySensor.getMaxThreshold() + 1;
                    humiditySensor.setMaxThreshold(mt);
                    if(mt >= maxMaxHygroThreshold) {
                        display(getResources().getText(R.string.max_hygro_reached).toString());
                    }
                }
                public void decreaseHygroMaxThreshold() {
                    int mt = humiditySensor.getMaxThreshold() - 1;
                    humiditySensor.setMaxThreshold(mt);
                    if(mt <= minMaxHygroThreshold) {
                        display(getResources().getText(R.string.min_hygro_reached).toString());
                    }
                }
                public int getThermoMaxThreshold() {
                    return temperatureSensor.getMaxThreshold();
                }
                public void increaseThermoMaxThreshold() {
                    int mt = temperatureSensor.getMaxThreshold() + 1;
                    temperatureSensor.setMaxThreshold(mt);
                    if(mt >= maxMaxThermoThreshold) {
                        display(getResources().getText(R.string.max_thermo_reached).toString());
                    }
                }
                public void decreaseThermoMaxThreshold() {
                    int mt = temperatureSensor.getMaxThreshold() - 1;
                    temperatureSensor.setMaxThreshold(mt);
                    if(mt <= minMaxThermoThreshold) {
                        display(getResources().getText(R.string.min_thermo_reached).toString());
                    }
                }
                public void kill() {
                    // not used
                }
            };

    @Override
    public void onTaskRemoved(Intent rootIntent) {
        Toast.makeText(this, "Task removed: " + rootIntent, Toast.LENGTH_LONG).show();
    }

    private static final int REPORT_MSG = 1;
    private static final int GD_MAX_HYGRO_MSG = 4;
    private static final int GD_MAX_THERMO_MSG = 5;
    private static final int ERROR_MSG = 6;

    /**
     * Our Handler used to execute operations on the main thread.  This is used
     * to schedule increments of our value.
     */
    private final Handler mHandler = new Handler() {
        @Override public void handleMessage(Message msg) {
            switch (msg.what) {

                case REPORT_MSG: {
                    // Broadcast to all clients the new value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).valueChanged(messageToBeSent);
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                        }
                    }
                    mCallbacks.finishBroadcast();

                } break;
                case GD_MAX_HYGRO_MSG: {
                    // Broadcast to all clients the max threshold value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).maxGdHygro(humiditySensor.getMaxThreshold() + "");
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                        }
                    }
                    mCallbacks.finishBroadcast();

                } break;
                case GD_MAX_THERMO_MSG: {
                    // Broadcast to all clients the max threshold value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i=0; i<N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).maxGdThermo(temperatureSensor.getMaxThreshold() + ":"+temperatureSensor.getMinThreshold());
                        } catch (RemoteException e) {
                            // The RemoteCallbackList will take care of removing
                            // the dead object for us.
                        }
                    }
                    mCallbacks.finishBroadcast();

                } break;
                case ERROR_MSG: {
                    // Broadcast to all clients the new value.
                    final int N = mCallbacks.beginBroadcast();
                    for (int i = 0; i < N; i++) {
                        try {
                            mCallbacks.getBroadcastItem(i).errorMessage(messageToBeSent);
                        } catch (RemoteException e) {
                            e.printStackTrace();
                        }
                    }
                    mCallbacks.finishBroadcast();
                } break;
                default:
                    super.handleMessage(msg);
            }
        }
    };

    public void runSample() {

        try {
            try {
                Log.d("DCD_DEBUG", "Before anything");
                SharedPreferences mSharedPref = getApplicationContext().getSharedPreferences(
                        getString(R.string.preference_file_key), Context.MODE_PRIVATE);

                // Create the directly-connected device instance
                if (mSharedPref.getBoolean(getString(R.string.use_provided_bks), false)) {
                    Log.d("DCD_DEBUG", "Using existing BKS");
                    gatewayDevice = new GatewayDevice(context);
                } else {
                    SharedPreferences pref = getApplicationContext().getSharedPreferences(
                        getString(R.string.preference_file_key), Context.MODE_PRIVATE);
                    try (FileOutputStream fos = openFileOutput(getString(R.string.last_known), Context.MODE_PRIVATE)) {
                        fos.write(pref.getString(getString(R.string.ta_file_path), "").getBytes());
                    } catch (IOException ex) {
                        Log.d(ERROR_TAG, "Can not safe to internal storage");
                    }
                    Log.d("DCD_DEBUG", "Using BKS" + mSharedPref.getString(getString(R.string.ta_file_path),"") + "---" + mSharedPref.getString(getString(R.string.ta_password),"") + " --------");

                    gatewayDevice = new GatewayDevice(mSharedPref.getString(getString(R.string.ta_file_path),""), mSharedPref.getString(getString(R.string.ta_password),""));
                }
            } catch (GeneralSecurityException dse) {
                // Tell the user
                displayError(dse.getMessage());
            }

            if (!gatewayDevice.isActivated()) {
                // If the device has not been activated, connect to the server
                // using client-credentials and activate the client to
                // obtain the private key. The private key is then persisted.
                gatewayDevice.activate();
            }

            temperatureSensor = new TemperatureSensor(gatewayDevice.getEndpointId() + "_Sample_TS");
            humiditySensor = new HumiditySensor(gatewayDevice.getEndpointId() + "_Sample_HS");

            // Get the device model instance.
            final DeviceModel temperatureDeviceModel =
                    gatewayDevice.getDeviceModel(TEMPERATURE_SENSOR_MODEL_URN);

            final DeviceModel humidityDeviceModel =
                    gatewayDevice.getDeviceModel(HUMIDITY_SENSOR_MODEL_URN);


            final String temperatureSensorEndpointId;
            final String humiditySensorEndpointId;

            Map<String, String> metaData = new HashMap<String, String>();
            metaData.put(
                    GatewayDevice.MANUFACTURER,
                    temperatureSensor.getManufacturer());
            metaData.put(
                    GatewayDevice.MODEL_NUMBER,
                    temperatureSensor.getModelNumber());
            metaData.put(
                    GatewayDevice.SERIAL_NUMBER,
                    temperatureSensor.getSerialNumber());

            temperatureSensorEndpointId =
                    gatewayDevice.registerDevice(
                            true, /* Default value for proposal. TODO: remove it after UI update */
                            temperatureSensor.getHardwareId(),
                            metaData,
                            TEMPERATURE_SENSOR_MODEL_URN);

            metaData = new HashMap<String, String>();
            metaData.put(
                    GatewayDevice.MANUFACTURER,
                    humiditySensor.getManufacturer());
            metaData.put(
                    GatewayDevice.MODEL_NUMBER,
                    humiditySensor.getModelNumber());
            metaData.put(
                    GatewayDevice.SERIAL_NUMBER,
                    humiditySensor.getSerialNumber());

            humiditySensorEndpointId =
                    gatewayDevice.registerDevice(
                            true, /* Default value for proposal. TODO: remove it after UI update */
                            humiditySensor.getHardwareId(),
                            metaData,
                            HUMIDITY_SENSOR_MODEL_URN);

            //
            // Create VirtualDevice instances to model the sensors
            // in the IoT Cloud Service
            //
            final VirtualDevice virtualizedTemperatureSensor =
                    gatewayDevice.createVirtualDevice(
                            temperatureSensorEndpointId,
                            temperatureDeviceModel);

            final VirtualDevice virtualizedHumiditySensor =
                    gatewayDevice.createVirtualDevice(
                            humiditySensorEndpointId,
                            humidityDeviceModel);

            //
            // For The temperatureSensor model, the min and max threshold values
            // can be written. Create a callback that will handle setting
            // the value on the temperature sensor device.
            //
            virtualizedTemperatureSensor.setOnChange(
                new VirtualDevice.ChangeCallback<VirtualDevice
                        >() {
                        @Override
                        public void onChange(VirtualDevice.ChangeEvent<VirtualDevice> event) {
                            VirtualDevice virtualDevice = event.getVirtualDevice();
                            VirtualDevice.NamedValue<?> namedValues = event.getNamedValue();
                            StringBuilder msg =
                                    new StringBuilder(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString());
                            msg.append(" : ");
                            msg.append(virtualDevice.getEndpointId());
                            msg.append(" : onChange : ");

                            boolean first = true;
                            for(VirtualDevice.NamedValue<?> namedValue = namedValues;
                                namedValue != null;
                                namedValue = namedValue.next()) {
                                final String attribute = namedValue.getName();
                                final Object value = namedValue.getValue();
                                if (!first) {
                                    msg.append(',');
                                } else {
                                    first = false;
                                }

                                if ("minThreshold".equals(attribute)) {
                                    int min = ((Integer)value).intValue();
                                    msg.append("\"minThreshold\"=");
                                    msg.append(min);
                                    temperatureSensor.setMinThreshold(min);
                                } else if ("maxThreshold".equals(attribute)) {
                                    int max = ((Integer)value).intValue();
                                    msg.append("\"maxThreshold\"=");
                                    msg.append(max);
                                    temperatureSensor.setMaxThreshold(max);
                                }
                            }

                            display(msg.toString());
                        }
                    }
            );

            //
            // The temperatureSensor model has a 'reset' action, which
            // resets the temperature sensor to factory defaults. Create
            // a callback that will handle calling reset on the temperature
            // sensor device.
            //
            virtualizedTemperatureSensor.setCallable("reset",
                    new VirtualDevice.Callable<Void>() {
                        @Override
                        public void call(VirtualDevice virtualDevice,
                                         Void not_used) {
                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() + " : " +
                                    virtualDevice.getEndpointId() +
                                " : Call : reset");

                            temperatureSensor.reset();
                            // After the sensor is reset, next poll of the
                            // sensor will yield new min and max temp values
                            // and update the values in the VirtualDevice
                        }
                    });

            //
            // The temperatureSensor model has a 'power' action, which
            // takes a boolean argument: true to power on the simulated device,
            // false to power off the simulated device. Create a callback that
            // will handle calling power on and power off on the temperature
            // sensor device.
            //
            virtualizedTemperatureSensor.setCallable("power",
                    new VirtualDevice.Callable<Boolean>() {
                        @Override
                        public void call(VirtualDevice virtualDevice,
                                         Boolean on) {
                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() + " : " +
                                    virtualDevice.getEndpointId() +
                                    " : Call : \"power\"=" + on);
                            boolean isOn = temperatureSensor.isPoweredOn();
                            if (on != isOn) {
                                temperatureSensor.power(on);
                            }

                            // Avoid re-entering the library, let the polling
                            // loop handle the attribute update.
                        }
                    });

            //
            // For the humiditySensor model, the maxThreshold attribute can be
            // written. Create a callback for setting the maxThreshold on the
            // humidity sensor device.
            //
            virtualizedHumiditySensor.setOnChange("maxThreshold",
                    new VirtualDevice.ChangeCallback<VirtualDevice>() {
                        @Override
                        public void onChange(VirtualDevice.ChangeEvent<VirtualDevice> event) {
                            VirtualDevice virtualDevice = event.getVirtualDevice();
                            VirtualDevice.NamedValue<?> namedValue = event.getNamedValue();
                            Integer value = (Integer)namedValue.getValue();
                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString()+ " : " +
                                virtualDevice.getEndpointId() +
                                " : onChange : \"maxThreshold\"=" + value);

                            humiditySensor.setMaxThreshold(value);
                        }
                    });


            //
            // Create a handler for errors that may be generated when trying
            // to set values on the virtual device. The same callback is used
            // for both virtual devices.
            //
            final VirtualDevice.ErrorCallback<VirtualDevice>
                errorCallback = new VirtualDevice.ErrorCallback<VirtualDevice
                    >() {
                @Override
                    public void onError(VirtualDevice.ErrorEvent event) {
                        VirtualDevice device =
                            (VirtualDevice)event.getVirtualDevice();
                        display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() + " : onError : " +
                            device.getEndpointId() +
                            " : \"" + event.getMessage());
                }
            };

            virtualizedTemperatureSensor.setOnError(errorCallback);
            virtualizedHumiditySensor.setOnError(errorCallback);

            //
            // Create an Alert for max/min temperature threshold and another
            // for maximum humidity threshold. Alerts do not have to be
            // created new each time an alerting event occurs.
            //
            final Alert tooHotAlert = virtualizedTemperatureSensor
                    .createAlert("urn:com:oracle:iot:device:temperature_sensor:too_hot");

            final Alert tooColdAlert = virtualizedTemperatureSensor
                    .createAlert("urn:com:oracle:iot:device:temperature_sensor:too_cold");

            final Alert tooHumidAlert = virtualizedHumiditySensor
                    .createAlert("urn:com:oracle:iot:device:humidity_sensor:too_humid");

            // Initialize the sensor to the device model's maxThreshold default value
            int defaultThreshold = virtualizedHumiditySensor.get("maxThreshold");
            humiditySensor.setMaxThreshold(defaultThreshold);
            rememberHygroMaxThreshold = humiditySensor.getMaxThreshold();

            // Initialize the sensor to the device model's min and
            // maxThreshold default value
            defaultThreshold = virtualizedTemperatureSensor.get("maxThreshold");
            temperatureSensor.setMaxThreshold(defaultThreshold);
            rememberThermoMaxThreshold = temperatureSensor.getMaxThreshold();
            defaultThreshold = virtualizedTemperatureSensor.get("minThreshold");
            temperatureSensor.setMinThreshold(defaultThreshold);

            sensorThread = new Thread(
                    new Runnable() {
                        public void run() {

                            // Only generate an alert when the threshold is crossed
                            boolean humidityAlerted = false;
                            boolean tempAlerted = false;
                            boolean wasOff = true;
                            long waitTime=0;
                            while(!stopped) {
                                try {
                                    Thread.sleep(waitTime);
                                    waitTime = 5000;

                                    int humidity =
                                            humiditySensor.getHumidity();
                                    int humidityThreshold =
                                            humiditySensor.getMaxThreshold();

                                    display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() + " : " +
                                            humiditySensorEndpointId +
                                            " : Set : \"humidity\"=" + humidity);

                                    virtualizedHumiditySensor.set("humidity",
                                            humidity);

                                    if (humidity > humidityThreshold) {
                                        if (!humidityAlerted) {
                                            humidityAlerted = true;
                                            Thread.sleep(1000);

                                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() +
                                                    " : " +
                                                virtualizedHumiditySensor.
                                                            getEndpointId() +
                                                    " : Alert : " +
                                                    "\"humidity\"=" + humidity);
                                            tooHumidAlert.set("humidity",
                                                    humidity).raise();
                                        }
                                    } else {
                                        humidityAlerted = false;
                                    }

                                    Thread.sleep(waitTime);

                                    if (!temperatureSensor.isPoweredOn()) {
                                        wasOff = true;
                                        continue;
                                    }

                                    if (wasOff) {
                                        wasOff = false;

                                        Date startTime =
                                                temperatureSensor.getStartTime();
                                        double temp =
                                                temperatureSensor.getTemp();
                                        String unit =
                                                temperatureSensor.getUnit();
                                        double minTemp =
                                                temperatureSensor.getMinTemp();
                                        double maxTemp =
                                                temperatureSensor.getMaxTemp();
                                        int minThreshold =
                                                temperatureSensor.getMinThreshold();
                                        int maxThreshold =
                                                temperatureSensor.getMaxThreshold();

                                        Thread.sleep(1000);

                                        String displayStr = new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(startTime).toString() + " : " +
                                            virtualizedTemperatureSensor.
                                                        getEndpointId() + " : Set : " +
                                            "\"temp\"=" + temp +
                                                " : \"power\"=true" +
                                                " : \"unit\"=" + unit +
                                                " : \"minTemp\"=" + minTemp +
                                                " : \"maxTemp\"=" + maxTemp;

                                        // After the sensor is powered on,
                                        // the attribute values in the
                                        // VirtualDevice need to be updated
                                        // to reflect the new values.
                                        virtualizedTemperatureSensor.update()
                                                .set("startTime", startTime)
                                                .set("temp", temp)
                                                .set("unit", unit)
                                                .set("minTemp", minTemp)
                                                .set("maxTemp", maxTemp);
                                        // If minThreshold is not the default, include it
                                        if (minThreshold != 0) {
                                            virtualizedTemperatureSensor.set("minThreshold", minThreshold);
                                            displayStr += " : \"minThreshold\"=" +
                                                    minThreshold;
                                        }
                                        // If maxThreshold is not the default, include it
                                        if (maxThreshold != 70) {
                                            virtualizedTemperatureSensor.set("maxThreshold", maxThreshold);
                                            displayStr += ":\"maxThreshold\"=" +
                                                    maxThreshold;
                                        }
                                        virtualizedTemperatureSensor.finish();

                                        display(displayStr);
                                        continue;
                                    }

                                    double prevMinTemp = temperatureSensor.getMinTemp();
                                    double prevMaxTemp = temperatureSensor.getMaxTemp();
                                    double temperature = temperatureSensor.getTemp();
                                    double minTemp = temperatureSensor.getMinTemp();
                                    double maxTemp = temperatureSensor.getMaxTemp();
                                    String id = temperatureSensorEndpointId;

                                    StringBuilder sb = new
                                            StringBuilder(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString());
                                    sb.append(" : ");
                                    sb.append(id);
                                    sb.append(" : Set : \"temp\"=");
                                    sb.append(temperature);
                                    virtualizedTemperatureSensor.update().set(
                                            "temp", temperature);

                                    if (minTemp != prevMinTemp) {
                                        sb.append(" : \"minTemp\"=");
                                        sb.append(minTemp);
                                        virtualizedTemperatureSensor.set(
                                                "minTemp", minTemp);
                                    }

                                    if (maxTemp != prevMaxTemp) {
                                        sb.append(" : \"maxTemp\"=");
                                        sb.append(maxTemp);
                                        virtualizedTemperatureSensor.set(
                                                "maxTemp", maxTemp);
                                    }
                                    Thread.sleep(1000);

                                    display(sb.toString());
                                    virtualizedTemperatureSensor.finish();

                                    String unit =
                                            temperatureSensor.getUnit();
                                    int minTempThreshold =
                                            temperatureSensor.getMinThreshold();
                                    int maxTempThreshold =
                                            temperatureSensor.getMaxThreshold();
                                    if (temperature > maxTempThreshold) {
                                        if (!tempAlerted) {
                                            Thread.sleep(1000);

                                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() +
                                                    " : " +
                                                    virtualizedTemperatureSensor.
                                                            getEndpointId() +
                                                    " : Alert : " +
                                                    "\"temp\"=" + temperature +
                                                    ":\"unit\"=" + unit +
                                                    ":\"maxThreshold\"=" +
                                                    maxTempThreshold);

                                            tooHotAlert
                                                    .set("temp", temperature)
                                                    .set("unit", unit)
                                                    .set("maxThreshold",
                                                            maxTempThreshold)
                                                    .raise();
                                        }
                                    } else if (temperature < minTempThreshold) {
                                        if (!tempAlerted) {
                                            Thread.sleep(1000);

                                            display(new SimpleDateFormat("HH:mm:ss", Locale.ROOT).format(new Date()).toString() +
                                                    " : " +
                                                virtualizedTemperatureSensor.
                                                            getEndpointId() +
                                                    " : Alert : " +
                                                    "\"temp\"=" + temperature +
                                                    ":\"unit\"=" + unit +
                                                    ":\"minThreshold\"=" +
                                                    minTempThreshold);
                                            tooColdAlert
                                                    .set("temp", temperature)
                                                    .set("unit", unit)
                                                    .set("minThreshold",
                                                            minTempThreshold)
                                                    .raise();
                                        }
                                    } else {
                                        tempAlerted = false;
                                    }
                                } catch (Exception e) {
                                    e.printStackTrace();
                                    if(e.getMessage() != null) {
                                        // Tell the user
                                        displayError(e.getMessage());
                                    } else {
                                        // Tell the user
                                        displayError("Error. No message.");
                                    }
                                }
                            }
                        }
                    });

            sensorThread.setDaemon(true);
            sensorThread.start();

        } catch (Exception e) {
            e.printStackTrace();
            Log.d("GD_ERROR", "runSample: -------------" + e.toString());
            if(e.getMessage() != null) {
                // Tell the user
                displayError(e.getMessage());
            } else {
                // Tell the user
                displayError("Error. No message.");
            }
        }
    }

    private void display(String s) {
        Log.d(MSG_TAG, s);
        messageToBeSent = s;
        messageSent = false;
        mHandler.sendEmptyMessage(REPORT_MSG);
        if(humiditySensor!=null) {
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }

            mHandler.sendEmptyMessage(GD_MAX_HYGRO_MSG);
        }
        if(temperatureSensor!=null) {
            try {
                Thread.sleep(2000);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
            mHandler.sendEmptyMessage(GD_MAX_THERMO_MSG);
        }
    }

    private void displayError(String string) {
        Log.d(ERROR_TAG, "Error Message: "+string);
        messageToBeSent = string;
        mHandler.sendEmptyMessage(ERROR_MSG);
        try {
            Thread.sleep(8000);
        } catch(InterruptedException e) {
            Thread.currentThread().interrupt();
        }
        System.exit(-1);
    }

    @Override
    public void onDestroy() {
        stopped = true;
        // power management: let the CPU go to sleep
        if (wakeLock != null) wakeLock.release();

        try {
            sensorThread.join();
        } catch (InterruptedException e) {
			Thread.currentThread().interrupt();
        }
        try {
            workerThread.join();
        } catch (InterruptedException e) {
			Thread.currentThread().interrupt();
        }
        humiditySensor.setMaxThreshold(rememberHygroMaxThreshold);
        temperatureSensor.setMaxThreshold(rememberThermoMaxThreshold);
        // Cancel the persistent notification.
        mNM.cancelAll();
        showNotification(false);
        // Tell the user we stopped.
        Toast.makeText(this, R.string.gd_service_stopped, Toast.LENGTH_SHORT).show();

        // Unregister all callbacks.
        mCallbacks.kill();

        mHandler.removeMessages(REPORT_MSG);
        mHandler.removeMessages(GD_MAX_HYGRO_MSG);
        mHandler.removeMessages(GD_MAX_THERMO_MSG);
    }

    /**
     * Show a notification while this service is running.
     */
    private void showNotification(boolean serviceStarted) {
        // In this sample, we'll use the same text for the ticker and the expanded notification
        CharSequence text = serviceStarted ? getText(R.string.gd_service_started) :
                getText(R.string.gd_service_stopped);
        // The PendingIntent to launch our activity if the user selects this notification
        PendingIntent contentIntent = PendingIntent.getActivity(this, 0,
                new Intent(this, MainActivity.class), 0);
        // Set the info for the views that show in the notification panel.
        Notification notification = new Notification.Builder(this)
                .setSmallIcon(R.drawable.mini)  // the status icon
                .setTicker(text)  // the status text
                .setWhen(System.currentTimeMillis())  // the time stamp
                .setContentTitle(getText(R.string.gd_service_label))  // the label of the entry
                .setContentText(text)  // the contents of the entry
                .setContentIntent(contentIntent)  // The intent to send when the entry is clicked
                .build();
        // Send the notification.
        // We use a string id because it is a unique number.  We use it later to cancel.
        mNM.notify(serviceStarted ? R.string.gd_service_started :
                R.string.gd_service_stopped, notification);
    }
}
